/*
Copyright (C)2006 George Ang (gnap.an AT gmail.com)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.
 
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
 
You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/ 

#include <kademlia/session.hpp>
#include <kademlia/utils/miscutils.hpp>

#ifndef FINDNODE_HPP
#define FINDNODE_HPP

namespace kademlia
{
	template<typename _uintX, class debugger = debug>
		class search_find_node: public session<_uintX, debugger>
	{
			typedef std::map<_uintX, contact<_uintX>* > contact_map;
			typedef kad<_uintX, debugger> kad_type;
			typedef typename contact_map::iterator contact_map_iterator;
		public:
			typedef _uintX uint_type;
			typedef contact<uint_type> contact_type;
			typedef std::list<contact_type*> contact_list;
			typedef typename std::list<contact<_uintX>* >::iterator contact_list_iterator;
			typedef typename std::list<contact<_uintX>* >::const_iterator contact_list_const_iterator;

			search_find_node();

			search_find_node(uint_type const& target, bool is_complete = false);

			~search_find_node();

			void start();

			void add_forwards(uint_type const& client_id, contact_list &list_results);

			void add_result(uint_type const&client_id, contact_type* contact_ptr);

			contact_type* get_result();

			void on_timer();

			bool is_finished();

			bool timed_out();

			void prepare_to_stop();

			enum
			{
				RUNNING = 1,
				FINISHED,
				STOPPING,
			};

		protected:
			boost::mutex m_mutex;
			bool m_is_complete_search;
			contact_map m_map_possible;
			contact_map m_map_tried;
			contact_map m_alpha_pool;
			contact_map m_map_in_use;
			
			contact_type *m_result;
			contact_list m_list_delete;
	};

	template<typename _uintX, class debugger>
		search_find_node<_uintX, debugger>::search_find_node()
		{
		}

	template<typename _uintX, class debugger>
		search_find_node<_uintX, debugger>::search_find_node(
				uint_type const& target,
				bool is_complete)
		{
			set_target_id(target);
			m_is_complete_search = is_complete;
			m_result = NULL;
		}

	template<typename _uintX, class debugger>
		search_find_node<_uintX, debugger>::~search_find_node()
		{
			boost::mutex::scoped_lock lock(m_mutex);
			for (contact_map_iterator iter = m_map_in_use.begin(); iter != m_map_in_use.end(); ++iter)
				iter->second->dec_use();

			for (contact_list_const_iterator iter = m_list_delete.begin(); iter != m_list_delete.end(); ++iter)
				delete *iter;
		}

	template<typename _uintX, class debugger>
		void search_find_node<_uintX, debugger>::start()
		{
			boost::mutex::scoped_lock lock(m_mutex);
			debugger::send(SESSION, "find node on target %s started.",this->m_target.get_hex_string().c_str());
			debugger::send(SESSION, "find node on target %s state = RUNNING.",this->m_target.get_hex_string().c_str());
			this->m_session_state = RUNNING;
			if (m_map_possible.empty())
				kad_type::get_routing_table()->get_closest_to(3, this->m_target, this->m_distance, 50, &m_map_possible, true, true);

			debugger::send(SESSION, "find node on target %s : %d requests on queque.",this->m_target.get_hex_string().c_str(),
					m_map_possible.size());

			if(!m_map_possible.empty())
			{
				for (contact_map_iterator iter = m_map_possible.begin(); iter != m_map_possible.end(); ++iter)
					m_map_in_use[iter->first] = iter->second;

		//		assert(m_map_possible.size() == m_map_in_use.size());
		//		assert(m_map_possible.size() != 50); //fix this later by George Ang
		// below was different from eMule's implementation. we start with alpha queries,and new forwards
		// will update alpha pool so we get a closer hop to the target. queries was controled by timers. 
		// 										by George Ang

				int count;

				count = min(ALPHA_QUERY, (int)m_map_possible.size());
				contact_map_iterator it_contact_map = m_map_possible.begin();

				for (int index = 0; index < count; index++)
				{
					contact_type *contact_ptr = it_contact_map->second;
					debugger::send(SESSION, "find node on target %s send request to %s.", this->get_target_id().get_hex_string().c_str(), contact_ptr->get_node_id().get_hex_string().c_str());
					m_map_tried[it_contact_map->first] = contact_ptr;
					kad_type::get_listener()->send_NODE_LOOKUP_REQUEST(this->get_target_id(), this->get_session_id(), contact_ptr);
					++it_contact_map;
				}
				
			}

		}

	template<typename _uintX, class debugger>
		void search_find_node<_uintX, debugger>::add_forwards(
				uint_type const& client_id,
				contact_list &list_results)
		{
			boost::mutex::scoped_lock lock(m_mutex);
			uint_type client_distance(client_id);
			client_distance._xor(this->get_target_id());
			this->m_last_response = time(NULL);
			if(m_map_tried.count(client_distance) <= 0)
				return;


			for (contact_list_iterator iter = list_results.begin();
					iter != list_results.end();
					++iter)
			{
				uint_type forward_distance((*iter)->get_node_id());
				forward_distance._xor(this->get_target_id());
				if(kad_type::get_listener()->add_contact(*iter, true))
				{
					(*iter)->inc_use();
					m_map_in_use[forward_distance] = *iter;
				}
				else
					m_list_delete.push_back(*iter);

				if(m_map_possible.count(forward_distance) > 0)
					continue;
				if(m_map_tried.count(forward_distance) > 0)
					continue;
				if(m_alpha_pool.count(forward_distance) > 0)
					continue;

				if(forward_distance < client_distance)
				{
					if(m_alpha_pool.size() < ALPHA_QUERY)
					{
						debugger::send(SESSION,
								"find node on target %s new alpha: %s.",
								this->m_target.get_hex_string().c_str(),
								(*iter)->get_node_id().get_hex_string().c_str());
						m_alpha_pool[forward_distance] = *iter;
					}
					else
						for(contact_map_iterator it_alpha = m_alpha_pool.begin();
								it_alpha != m_alpha_pool.end();
								it_alpha++)
						{
							if(forward_distance < it_alpha->first)
							{
								debugger::send(SESSION,
										"find node on target %s new alpha: %s.",
										this->m_target.get_hex_string().c_str(),
										(*iter)->get_node_id().get_hex_string().c_str());
								m_alpha_pool.erase(it_alpha);
								m_alpha_pool[forward_distance] = *iter;
							}
						}
				}
			}

		}

	template<typename _uintX, class debugger>
		void search_find_node<_uintX, debugger>::add_result(
				uint_type const& client_id,
				contact_type* contact_ptr)
		{
			debugger::send( SESSION,
					"find node on target %s reached.",
					this->m_target.get_hex_string().c_str(),
					client_id.get_hex_string().c_str());

			this->m_last_response = time(NULL);
			m_result = contact_ptr;
		}

	template<typename _uintX, class debugger>
		class search_find_node<_uintX, debugger>::contact_type* search_find_node<_uintX, debugger>::get_result()
		{
			return m_result;
		}

	template<typename _uintX, class debugger>
		void search_find_node<_uintX, debugger>::on_timer()
		{
			boost::mutex::scoped_lock lock(m_mutex);
			if (this->m_session_state == RUNNING)
			{
				if ( m_result )
				{
					if (!m_is_complete_search)
					{
						debugger::send(SESSION,
								"find node on target %s state RUNNING->FINISHED.",
								this->m_target.get_hex_string().c_str());
						this->m_session_state = FINISHED;
					}
					else
						prepare_to_stop();
					return;
				}

				if (m_map_possible.empty() && m_alpha_pool.empty() )
				{
					debugger::send(SESSION,
							"find node on target %s state RUNNING->FINISHED.",
							this->m_target.get_hex_string().c_str());
					this->m_session_state = FINISHED;
					return;
				}



				if((this->m_last_response + 3) > time(NULL))
					return;

				for(contact_map_iterator iter = m_alpha_pool.begin();
						iter != m_alpha_pool.end();
						++iter)
				{
					contact_type* contact_ptr = iter->second;

					if (m_map_tried.count(iter->first) > 0)
						continue;
					else
					{
						m_map_tried[iter->first] = contact_ptr;
						debugger::send(SESSION,
								"find node on target %s send closer hops to %s.",
								this->m_target.get_hex_string().c_str(),
								contact_ptr->get_node_id().get_hex_string().c_str());
						kad_type::get_listener()->send_NODE_LOOKUP_REQUEST(this->m_target,
								this->get_session_id(),
								contact_ptr);
					}
				}

				while(!m_map_possible.empty())
				{
					contact_type* contact_ptr = m_map_possible.begin()->second;

					if (m_map_tried.count(m_map_possible.begin()->first) > 0)
					{

						if(m_map_possible.count(m_map_possible.begin()->first) > 0)

							m_map_possible.erase(m_map_possible.begin());
					}
					else
					{
						m_map_tried[m_map_possible.begin()->first] = contact_ptr;
						debugger::send(SESSION,
								"find node on target %s send request to %s.",
								this->m_target.get_hex_string().c_str(),
								contact_ptr->get_node_id().get_hex_string().c_str());

						kad_type::get_listener()->send_NODE_LOOKUP_REQUEST(
								this->m_target,
								this->get_session_id(),
								contact_ptr);
						return;
					}
				}
			}

		}

	template<typename _uintX, class debugger>
		bool search_find_node<_uintX, debugger>::is_finished()
		{
			return this->m_session_state == FINISHED;
		}

	template<typename _uintX, class debugger>
		bool search_find_node<_uintX, debugger>::timed_out()
		{
			time_t now = time(NULL);

			return (this->m_created + SEARCHNODE_LIFETIME) < now;
		}

	template<typename _uintX, class debugger>
		void search_find_node<_uintX, debugger>::prepare_to_stop()
		{
			if (this->m_session_state == STOPPING)
				return;

			this->m_created = time(NULL) - SEARCHNODE_LIFETIME + 15;
			debugger::send(SESSION,
					"find node on target %s state =STOPPING.",
					this->m_target.get_hex_string().c_str());
			this->m_session_state = STOPPING;
	}

};

#endif //FINDNODE_HPP
