/*
Ported by George Ang<gnap.an<AT>gmail.com>
Copyright (C)2003 Barry Dunne (http://www.emule-project.net)
 
This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.
 
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
 
You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/ 

#include <map>
#include <memory>
#include <boost/thread/thread.hpp>
#include <boost/thread/mutex.hpp>
#include <boost/thread/xtime.hpp>
#include <boost/bind.hpp>
#include <kademlia/routingzone.hpp>
#include <kademlia/contact.hpp>
#include <kademlia/kbucket.hpp>
#include <kademlia/session.hpp>
#include <kademlia/storage.hpp>
#include <kademlia/sessionmanager.hpp>
#include <kademlia/utils/uint160.hpp>
#include <kademlia/utils/debug.hpp>
#include <kademlia/kadlistener.hpp>

#ifndef KADEMLIA_HPP
#define KADEMLIA_HPP


namespace kademlia
{
	template<typename _uintX = uint160, class debugger = debug>
	class kad
	{
		public:
			typedef _uintX uint_type;
			typedef contact<uint_type> contact_type;
			typedef kad<uint_type, debugger> kad_type;
			typedef routing_zone<uint_type, debugger> routing_zone_type;
			typedef kad_listener<uint_type, debugger> kad_listener_type;
			typedef session_manager<uint_type, debugger> session_manager_type;
			typedef storage<uint_type, debugger> storage_type;
		private:
			typedef std::map<routing_zone_type*, routing_zone_type*> event_map;
			typedef typename event_map::const_iterator event_map_const_iterator;

		public:
			kad();

			kad(uint_type const& kad_id);

			static void start();

			static void stop();

			static uint_type get_kad_id();

			static routing_zone_type *get_routing_table();

			static void set_listener(kad_listener_type *);
			static kad_listener_type *get_listener();

			static void set_storage( storage_type *);
			static storage_type *get_storage();

			static bool is_running();

//			static bool	is_connected();
//			static void	bootstrap(contact_type *contact_ptr);
			
			static void add_event(routing_zone_type *zone_ptr);

			static void remove_event(routing_zone_type *zone_ptr);

			static void internal();
		private:
			static void process();

			static kad_type *m_instance;
			static event_map	m_map_events;
			static time_t m_next_session_timer;
			static time_t m_next_self_lookup;
			static time_t m_big_timer;
			static time_t m_bootstrap;
			static time_t m_consolidate;
			static uint32_t m_last_contact;
			static bool	m_running;
			static uint_type  m_kad_id;
			static std::auto_ptr<boost::thread> m_process_trd;
			static boost::mutex m_event_mutex;
			routing_zone_type *m_routing_zone;
			kad_listener_type *m_listener;
			storage_type *m_storage;
	};
	template<typename _uintX, class debugger>
		kad<_uintX, debugger>::kad()
		{
		}

	template<typename _uintX, class debugger>
		kad<_uintX, debugger>::kad(uint_type const& kad_id)
		{
			m_kad_id = kad_id;
		}

	template<typename _uintX, class debugger>
		void kad<_uintX, debugger>::start()
		{
			try
			{
				// If we already have a instance, something is wrong. 
				if( m_instance )
				{
					debugger::send(PROCESS,"alredy runing");
					return;
				}


				debugger::send(PROCESS,"starting kademlia");

				// Init session timer.
				m_next_session_timer = time(NULL);
				// Force a find_nodeComplete within the first 3 minutes.
				m_next_self_lookup = time(NULL) + 180;
				// Init big timer for Zones
				m_big_timer = time(NULL);
				// Init contact consolidate timer;
				m_consolidate = time(NULL) + 2700;
				// Init our random seed.
				srand((int)time(NULL));
				if(m_kad_id == uint_type(false))
					m_kad_id = uint_type(uint_type(false), 0);

				debugger::send(PROCESS ,
						"running with client id :%s",
						m_kad_id.get_hex_string().c_str());

				// Create our Kad objects.
				m_instance = new kad();
				m_instance->m_routing_zone = NULL;
				m_instance->m_routing_zone = new routing_zone_type();
	/*			
				if(m_instance->m_listener == NULL)
				{
					if(listener_ptr == NULL)
						m_instance->m_listener = new udp_listener<uint_type, debugger>();
					else
						m_instance->m_listener = listener_ptr;
				}
				if(m_instance->m_storage == NULL)
				{
					if(storage_ptr == NULL)
						m_instance->m_storage = new mem_storage<uint_type, debugger>();
					else
						m_instance->m_storage = storage_ptr;
				}
				m_instance->m_listener = listener_ptr;
				m_instance->m_storage = storage_ptr;
				*/
				// Mark Kad as running state.

				m_running = true;
				m_process_trd = std::auto_ptr<boost::thread>(new boost::thread(boost::bind(process)));
			}
			catch (...)
			{
				debugger::send(PROCESS, "error occured during starting");
			}
		}
	template<typename _uintX, class debugger>
		void kad<_uintX, debugger>::stop()
		{
			// Make sure we are running to begin with.
			if( !m_running )
				return;

			debugger::send(PROCESS ,"stopping ...");

			// Mark Kad as being in the stop state to make sure nothing else is used.
			m_running = false;

			// Remove all active searches.
			session_manager_type::stop_all_sessions();

			// Delete all Kad Objects.
		//	delete m_instance->m_listener;
			m_instance->m_listener = NULL;

			delete m_instance->m_routing_zone;
			m_instance->m_routing_zone = NULL;


		//	delete m_instance->m_pf;
		//	m_instance->m_pf = NULL;

			delete m_instance;
			m_instance = NULL;

			boost::mutex::scoped_lock lock(m_event_mutex);
			// Make sure all zones are removed.
			m_map_events.clear();

		}
	template<typename _uintX, class debugger>
		typename kad<_uintX, debugger>::uint_type kad<_uintX, debugger>::get_kad_id()
		{
			return m_kad_id;
		}

	template<typename _uintX, class debugger>
		typename kad<_uintX, debugger>::routing_zone_type* kad<_uintX, debugger>::get_routing_table()
		{
			if (m_instance == NULL || m_instance->m_routing_zone == NULL)
			{
				assert(0);
				return NULL;
			}
			return m_instance->m_routing_zone;

		}

	template<typename _uintX, class debugger>
		typename kad<_uintX, debugger>::kad_listener_type* kad<_uintX, debugger>::get_listener()
		{
			if (m_instance == NULL || m_instance->m_listener == NULL)
			{
				assert(0);
				return NULL;
			}
			return m_instance->m_listener;
		}

	template<typename _uintX, class debugger>
		void kad<_uintX, debugger>::set_listener(kad_listener_type * listener)
		{
			if (m_instance == NULL )
			{
				assert(0);
				return;
			}

			if (m_instance->m_listener == NULL)
				delete m_instance->m_listener;

			m_instance->m_listener= listener;
			
			return;
		}

	template<typename _uintX, class debugger>
		typename kad<_uintX, debugger>::storage_type* kad<_uintX, debugger>::get_storage()
		{
			if (m_instance == NULL || m_instance->m_storage == NULL)
			{
				assert(0);
				return NULL;
			}
			return m_instance->m_storage;
		}

	template<typename _uintX, class debugger>
		void kad<_uintX, debugger>::set_storage(storage_type * storage)
		{
			if (m_instance == NULL )
			{
				assert(0);
				return;
			}

			if (m_instance->m_storage == NULL)
				delete m_instance->m_storage;

			m_instance->m_storage= storage;
			
			return;
		}

	template<typename _uintX, class debugger>
		bool kad<_uintX, debugger>::is_running()
		{
			return m_running;
		}

	template<typename _uintX, class debugger>
		void kad<_uintX, debugger>::add_event(routing_zone_type *zone_ptr)
		{
			boost::mutex::scoped_lock lock(m_event_mutex);
			m_map_events[zone_ptr] = zone_ptr;
		}

	template<typename _uintX, class debugger>
		void kad<_uintX, debugger>::remove_event(routing_zone_type *zone_ptr)
		{
			boost::mutex::scoped_lock lock(m_event_mutex);
			m_map_events.erase(zone_ptr);
		}

	template<typename _uintX, class debugger>
		void kad<_uintX, debugger>::internal()
		{
			if( m_instance == NULL || !m_running)
			{
				debugger::send(PROCESS, "no instance or not runing.");
			}

			time_t now = time(NULL);
			if (m_next_self_lookup <= now)
			{
				session_manager_type::find_node(m_kad_id, true);
				m_next_self_lookup = 14400 + now; // 3600*4 seconds = 4 hours
			}

			for (event_map_const_iterator itevent_map = m_map_events.begin();
					itevent_map != m_map_events.end();
					++itevent_map)
			{
				boost::mutex::scoped_lock lock(m_event_mutex);
				routing_zone_type* zone_ptr = (routing_zone_type *)itevent_map->first;

				if (m_big_timer <= now)
				{
					if( zone_ptr->m_next_big_timer <= now )
					{
						if(zone_ptr->on_big_timer())
						{
							zone_ptr->m_next_big_timer = 3600 + now; //1 hour
							m_big_timer = 10 + now;
						}
					}
					else
					{
						if( m_last_contact && ( (now - m_last_contact) > (1200- 300)))  // 20min - 5 min
						{
							if(zone_ptr->on_big_timer())
							{
								zone_ptr->m_next_big_timer = 3600 + now; // 1 hour
								m_big_timer = 10 + now;
							}
						}
					}
				}
				if (zone_ptr->m_next_small_timer <= now)
				{
					zone_ptr->on_small_timer();
					zone_ptr->m_next_small_timer = 60 + now; // 1 min
				}
			}

			// This is a convenient place to add this, although not related to routing
			if (m_next_session_timer <= now)
			{
				session_manager_type::on_timer();
				m_next_session_timer = SESSION_TICKS + now;
			}

			// Try to consolidate any zones that are close to empty.
			if (m_consolidate <= now)
			{
				uint32_t merged_count = m_instance->m_routing_zone->consolidate();
				if(merged_count)
				m_consolidate = 2700 + now; // 45 min
			}
		}

	template<typename _uintX, class debugger>
		void kad<_uintX, debugger>::process()
		{
			while(1)
			{
				boost::xtime xt;
				boost::xtime_get(&xt, boost::TIME_UTC);
				xt.sec += 1;
				boost::thread::sleep(xt);
				internal();
			}
		}

	template<typename _uintX, class debugger>
		typename kad<_uintX, debugger>::kad_type* kad<_uintX, debugger>::m_instance = NULL;
	template<typename _uintX, class debugger>
		typename kad<_uintX, debugger>::event_map kad<_uintX, debugger>::m_map_events;
	template<typename _uintX, class debugger>
		time_t kad<_uintX, debugger>::m_next_session_timer;
	template<typename _uintX, class debugger>
		time_t kad<_uintX, debugger>::m_next_self_lookup;
	template<typename _uintX, class debugger>
		time_t kad<_uintX, debugger>::m_big_timer;
	template<typename _uintX, class debugger>
		time_t kad<_uintX, debugger>::m_consolidate;
	template<typename _uintX, class debugger>
		time_t kad<_uintX, debugger>::m_bootstrap;
	template<typename _uintX, class debugger>
		bool kad<_uintX, debugger>::m_running = false;
	template<typename _uintX, class debugger>
		_uintX kad<_uintX, debugger>::m_kad_id = _uintX(false) ;
	template<typename _uintX, class debugger>
		uint32_t kad<_uintX, debugger>::m_last_contact = 0;
	template<typename _uintX, class debugger>
		std::auto_ptr<boost::thread> kad<_uintX, debugger>::m_process_trd;
	template<typename _uintX, class debugger>
		boost::mutex kad<_uintX, debugger>::m_event_mutex;
}

#endif //KADEMLIA_HPP
