/*
Ported by George Ang<gnap.an<AT>gmail.com>
Copyright (C)2003 Barry Dunne (http://www.emule-project.net)
 
This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.
 
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
 
You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/ 

#include <string>
#include <math.h>
#include <list>
#include <map>
#include <boost/thread/mutex.hpp>
#include <kademlia/defines.hpp>
#include <kademlia/utils/debug.hpp>
#include <kademlia/contact.hpp>

#ifndef ROUTETABLE_HPP
#define ROUTETABLE_HPP

namespace kademlia
{
	template<typename _uintX, class debugger>class kad;
	template<typename _uintX, class debugger>class session_manager;
	template<typename _uintX, class debugger> class kbucket;
	/**
	 * The *Zone* is just a node in a binary tree of *Zone*s.
	 * Each zone is either an internal node or a leaf node.
	 * Internal nodes have "bin == null" and "subZones[i] != null",
	 * leaf nodes have "subZones[i] == null" and "bin != null".
	 * 
	 * All key pseudoaddresses are relative to the center (self), which
	 * is considered to be 000..000
	 */
	template<typename _uintX, class debugger = debug>
	class routing_zone
	{
		public:
			typedef _uintX uint_type;
			typedef contact<uint_type> contact_type;
			typedef std::list<contact_type*> contact_list;
			typedef routing_zone<uint_type, debugger> routing_zone_type;
			typedef typename std::list<contact<_uintX>* >::iterator contact_list_iterator;
			typedef typename std::list<contact<_uintX>* >::const_iterator contact_list_const_iterator;
			typedef std::map<uint_type, contact_type* > contact_map;

			routing_zone();

			~routing_zone();
			
			uint32_t consolidate();
			
			bool on_big_timer();

			void on_small_timer();

			bool add_contact(contact_type *contact_ptr);

			bool set_alive(contact_type *contact_ptr);

			size_t get_num_contacts() const;

			// Returns a list of all contacts in all leafs of this zone.
			contact_type *get_contact(uint_type const& node_id) const;

			void get_all_nodes(contact_list *list_result, bool empty_first = true);

			// Returns the *maxRequired* tokens that are closest to the target within this zone's subtree.
			uint32_t get_closest_to(
					uint8_t max_type,
				       	const uint_type &target,
				       	const uint_type &distance,
				       	uint32_t max_required,
				       	contact_map *result_map, 
					bool empty_first =  true,
				       	bool set_in_use = false) const;

			size_t get_bootstrap_contacts(contact_list *list_result, size_t max_required);

			uint32_t estimate_count();

			routing_zone_type* bottom();

			time_t m_next_big_timer;
			time_t m_next_small_timer;
		private:
			routing_zone(routing_zone_type *super_zone, int level, uint_type const&zone_index);

			void init(routing_zone_type *super_zone, int level, uint_type const&zone_index);

			bool is_leaf() const;

			bool can_split() const;

			// Returns all contacts from this zone tree that are no deeper than *depth* from the current zone.
			void top_depth(int depth, contact_list *list_result, bool empty_first = true);

			// Returns the maximum depth of the tree as the number of edges of the longest path to a leaf.
			size_t get_max_depth() const;

			void random_bucket(contact_list *list_result, bool empty_first = true);

			void split();

			void start_timer();

			void stop_timer();

			void random_lookup();

			/**
			* Generates a new TokenBin for this zone. Used when the current zone is becoming a leaf zone.
			* Must be deleted by caller
			*/
			routing_zone_type* gen_subzone(int side);

			/**
			* Zone pair is an array of two. Either both entries are null, which
			* means that *this* is a leaf zone, or both are non-null which means
			* that *this* has been split into equally sized finer zones.
			* The zone with index 0 is the one closer to our *self* token.
			*/
			routing_zone_type* m_subzones[2];
			routing_zone_type* m_superzone;
			static uint_type m_kad_id;
			/**
			* The level indicates what size chunk of the address space
			* this zone is representing. Level 0 is the whole space,
			* level 1 is 1/2 of the space, level 2 is 1/4, etc.
			*/
			uint32_t m_level;
			/**
			* This is the distance in number of zones from the zone at this level
			* that contains the center of the system; distance is wrt the XOR metric.
			*/
			uint_type m_zone_index;


			boost::mutex m_mutex;
			/** List of contacts, if this zone is a leaf zone. */
			kbucket<uint_type, debugger> *m_bucket;

	};
	template<typename _uintX, class debugger>
		routing_zone<_uintX, debugger>::routing_zone()
		{
			m_kad_id = kad<uint_type, debugger>::get_kad_id();

			debugger::send(PROCESS, "initilazing routing table...");
			debugger::send(PROCESS,
					"base XOR metric of %s",
					kad<uint_type,
					debugger>::get_kad_id().get_hex_string().c_str());

			init(NULL, 0, uint_type((uint32_t)0));
		}

	template<typename _uintX, class debugger>
		routing_zone<_uintX, debugger>::~routing_zone()
		{
			boost::mutex::scoped_lock lock(m_mutex);
			debugger::send(PROCESS,
					"zone level %d, index %s destroyed.",
					m_level,
					m_zone_index.get_binary_string(true).c_str());
			// If this zone is a leaf, delete our contact bucket.
			if (is_leaf())
				delete m_bucket;
			else
			{
				// If this zone is branch, delete it's leafs.
				delete m_subzones[0];
				delete m_subzones[1];
			}

		}

	template<typename _uintX, class debugger>
		uint32_t routing_zone<_uintX, debugger>::consolidate()
		{
			boost::mutex::scoped_lock lock(m_mutex);
			uint32_t merge_count = 0;
			if (is_leaf() )
				return merge_count;
			assert(m_bucket == NULL);
			if( !m_subzones[0]->is_leaf() )
				merge_count += m_subzones[0]->consolidate();
			if( !m_subzones[1]->is_leaf() )
				merge_count += m_subzones[1]->consolidate();
			if( m_subzones[0]->is_leaf() && m_subzones[1]->is_leaf() && get_num_contacts() < K/2 )
			{
				m_bucket = new kbucket<uint_type, debugger>();
				m_subzones[0]->stop_timer();
				m_subzones[1]->stop_timer();
				if(get_num_contacts() > 0)
				{
					contact_list list0;
					contact_list list1;
					m_subzones[0]->m_bucket->get_all_nodes(&list0);
					m_subzones[1]->m_bucket->get_all_nodes(&list1);

					for(contact_list_const_iterator iter = list0.begin(); 
							iter != list0.end();
							++iter)
						m_bucket->add_contact(*iter);

					for(contact_list_const_iterator iter = list1.begin();
							iter != list1.end();
							++iter)
						m_bucket->add_contact(*iter);
				}

				m_subzones[0]->m_superzone = NULL;
				m_subzones[1]->m_superzone = NULL;
				m_subzones[0]->m_bucket->m_dont_delete_contacts = true;
				m_subzones[1]->m_bucket->m_dont_delete_contacts = true;
				delete m_subzones[0];
				delete m_subzones[1];
				m_subzones[0] = NULL;
				m_subzones[1] = NULL;
				start_timer();
				merge_count++;
			}
			debugger::send(ROUTETABLE,
					"zone level %d, index %s consolidate : merged %u zones",
					m_level,
					m_zone_index.get_binary_string(true).c_str(),
					merge_count);

			return merge_count;

		}

	template<typename _uintX, class debugger>
		bool routing_zone<_uintX, debugger>::on_big_timer()
		{
			if (is_leaf()
				&&(m_zone_index < KK
					|| m_level < KBASE
					|| m_bucket->get_remaining() >= (K*.8)))
			{
				random_lookup();
				return true;
			}

			return false;

		}

	template<typename _uintX, class debugger>
		void routing_zone<_uintX, debugger>::on_small_timer()
		{
			boost::mutex::scoped_lock lock(m_mutex);
			if (!is_leaf())
				return;

			contact_type *contact_ptr = NULL;
			time_t now = time(NULL);
			contact_list list_nodes;
			// Remove dead entries
			m_bucket->get_all_nodes(&list_nodes);
			for(contact_list_iterator iter = list_nodes.begin();
					iter != list_nodes.end();
					++iter)
			{
				contact_ptr = *iter;
				if(contact_ptr->get_type() == 4)
				{
					if (((contact_ptr->get_expire_time() > 0)
						&& (contact_ptr->get_expire_time() <= now)))
					{
						if(!contact_ptr->in_use())
						{
							debugger::send(ROUTETABLE,
									"zone level %d, index %s contact %s seems too old, delete it.", 
									m_level, 
									m_zone_index.get_binary_string(true).c_str(),
									contact_ptr->get_node_id().get_hex_string().c_str());
							m_bucket->remove_contact(contact_ptr);
							delete contact_ptr;
						}
						continue;
					}
				}
				if(contact_ptr->get_expire_time() == 0)
					contact_ptr->set_expire_time(now);
			}
			contact_ptr = m_bucket->get_oldest();
			if(contact_ptr != NULL)
			{
				if(contact_ptr->get_expire_time() >= now
					|| contact_ptr->get_type() == 4)
				{
					m_bucket->remove_contact(contact_ptr);
					m_bucket->m_nodes.push_back(contact_ptr);
					contact_ptr = NULL;
				}
			}
			if(contact_ptr != NULL)
			{
				debugger::send(ROUTETABLE,
						"send kademlia HELLO to client %s. ",
						contact_ptr->get_node_id().get_hex_string().c_str());
				contact_ptr->checking_type();
				kad<uint_type, debugger>::get_listener()->send_HELLO(contact_ptr);
			}

		}

	template<typename _uintX, class debugger>
		bool routing_zone<_uintX, debugger>::add_contact(contact_type *contact_ptr)
		{
			// If we are not a leaf, call add on ther correct branch.
			if (! contact_ptr->is_valid() )
			{
				debugger::send(ROUTETABLE,
						"zone level %d, index %s contact %s not valid.", 
						m_level, 
						m_zone_index.get_binary_string(true).c_str(),
						contact_ptr->get_node_id().get_hex_string().c_str());
				return false;
			}


			if (!is_leaf())
				return m_subzones[contact_ptr->get_distance().get_bit_number(m_level)]->add_contact(contact_ptr);
			else
			{
				// Do we already have a contact with this KadID?
				if (set_alive(contact_ptr))
				{
					debugger::send(ROUTETABLE,
							"zone level %d, index %s contact %s refreshed.", 
							m_level, 
							m_zone_index.get_binary_string(true).c_str(),
							contact_ptr->get_node_id().get_hex_string().c_str());
					return true;
				}
				else if (m_bucket->get_remaining())
				{
					// This bucket is not full, so add the new contact.
					 boost::mutex::scoped_lock lock(m_mutex);
					if(m_bucket->add_contact(contact_ptr))
					{
						debugger::send(ROUTETABLE,
								"zone level %d, index %s new contact %s added.", 
								m_level, 
								m_zone_index.get_binary_string(true).c_str(),
								contact_ptr->get_node_id().get_hex_string().c_str(),
								contact_ptr);
						return true;
					}
					debugger::send(ROUTETABLE,
							"zone level %d, index %s new contact %s add refused.", 
							m_level, 
							m_zone_index.get_binary_string(true).c_str(),
							contact_ptr->get_node_id().get_hex_string().c_str());

					return false;
				}
				else if (can_split())
				{
					// This bucket was full and split, call add on the correct branch.
					split();
					return m_subzones[contact_ptr->get_distance().get_bit_number(m_level)]->add_contact(contact_ptr);
				}
				else
					return false;
			}

		}

	template<typename _uintX, class debugger>
		bool routing_zone<_uintX, debugger>::set_alive(contact_type *contact_ptr)
		{
			if (is_leaf())
				return m_bucket->set_alive(contact_ptr);
			else
			{
				if (m_subzones[0]->set_alive(contact_ptr))
					return true;

				return m_subzones[1]->set_alive(contact_ptr);
			}

		}

	template<typename _uintX, class debugger>
		size_t routing_zone<_uintX, debugger>::get_num_contacts() const
		{
			if (is_leaf())
				return m_bucket->get_size();
			else
				return m_subzones[0]->get_num_contacts() + m_subzones[1]->get_num_contacts();

		}

		// Returns a list of all contacts in all leafs of this zone.
	template<typename _uintX, class debugger>
		typename routing_zone<_uintX, debugger>::contact_type*
		routing_zone<_uintX, debugger>::get_contact(uint_type const& node_id) const
		{
			if (is_leaf())
				return m_bucket->get_contact(node_id);
			else
				return m_subzones[node_id.get_bit_number(m_level)]->get_contact(node_id);

		}

	template<typename _uintX, class debugger>
		void routing_zone<_uintX, debugger>::get_all_nodes(contact_list *list_result, bool empty_first)
		{
			if (is_leaf())
				m_bucket->get_all_nodes(list_result, empty_first);
			else
			{
				m_subzones[0]->get_all_nodes(list_result, empty_first);
				m_subzones[1]->get_all_nodes(list_result, false);
			}

		}

		// Returns the *maxRequired* tokens that are closest to the target within this zone's subtree.
	template<typename _uintX, class debugger>
		uint32_t routing_zone<_uintX, debugger>::get_closest_to(
				uint8_t max_type,
				uint_type const& target,
				uint_type const& distance,
				uint32_t max_required,
				contact_map *result_map, 
				bool empty_first,
				bool set_in_use) const
		{
			// If leaf zone, do it here
			if (is_leaf())
				return m_bucket->get_closest_to(
						max_type,
						target, 
						max_required,
						result_map,
						empty_first,
						set_in_use);
			// otherwise, recurse in the closer-to-the-target subzone first
			int closer = distance.get_bit_number(m_level);
			uint32_t found = m_subzones[closer]->get_closest_to(
					max_type,
					target,
					distance,
					max_required,
					result_map,
					empty_first,
					set_in_use);
			// if still not enough tokens found, recurse in the other subzone too
			if (found < max_required)
				found += m_subzones[1-closer]->get_closest_to(
						max_type,
						target,
						distance,
						max_required,
						result_map,
						false,
						set_in_use);
			return found;

		}

	template<typename _uintX, class debugger>
		size_t routing_zone<_uintX, debugger>::get_bootstrap_contacts(
				contact_list *list_result,
				size_t max_required)
		{
			boost::mutex::scoped_lock lock(m_mutex);
			assert(m_superzone == NULL);
			list_result->clear();
			uint32_t ret_val = 0;
			try
			{
				contact_list top;
				top_depth(LOG_BASE_EXPONENT, &top);
				if (top.size() > 0)
				{
					for(contact_list_const_iterator iter = top.begin();
							iter != top.end();
							++iter)
					{
						list_result->push_back(*iter);
						ret_val++;
						if(ret_val == max_required)
							break;
					}
				}
			}
			catch(...)
			{
				debugger::send(ROUTETABLE, "exception in routing_zone::get_booststrap_contacst");
			}
			return ret_val;

		}

	template<typename _uintX, class debugger>
		uint32_t routing_zone<_uintX, debugger>::estimate_count()
		{
			boost::mutex::scoped_lock lock(m_mutex);
			if (!is_leaf())
				return 0;
			if (m_level < KBASE)
				return (uint32_t)(pow(2, m_level)*K);
			routing_zone<uint_type, debugger>* cur_zone = m_superzone->m_superzone->m_superzone;
			// Find out how full this part of the tree is.
			float modify = ((float)cur_zone->get_num_contacts())/(float)(K*2);
			// First calculate users assuming the tree is full.
			// Modify count by bucket size.
			// Modify count by how full the tree is.
			// Modify count by assuming 20% of the users are firewalled and can't be a contact.
			return (uint32_t)((pow(2, m_level-2))*(float)K*modify*(1.20F));

		}

	template<typename _uintX, class debugger>
		routing_zone<_uintX, debugger>* routing_zone<_uintX, debugger>::bottom()
		{
			boost::mutex::scoped_lock lock(m_mutex);
			if(is_leaf())
				return this;
			return m_subzones[0]->bottom();
		}

	template<typename _uintX, class debugger>
		routing_zone<_uintX, debugger>::routing_zone(routing_zone_type* super_zone, int level, uint_type const& zone_index)
		{
			//Creat a new leaf.
			init(super_zone, level, zone_index);
		}

	template<typename _uintX, class debugger>
		void routing_zone<_uintX, debugger>::init(routing_zone_type* super_zone, int level, uint_type const& zone_index)
		{
			boost::mutex::scoped_lock lock(m_mutex);
			// Init all Zone vars
			// Set this zones parent
			m_superzone = super_zone;
			// Set this zones level
			m_level = level;
			// Set this zones uint160 index
			m_zone_index = zone_index;

			debugger::send(ROUTETABLE,
					"init zone level %d, index %s",
					m_level,
					m_zone_index.get_binary_string(true).c_str());
			// Mark this zone has having now leafs.
			m_subzones[0] = NULL;
			m_subzones[1] = NULL;
			// Create a new contact bucket as this is a leaf.
			m_bucket = new kbucket<uint_type, debugger>();

			// Set timer so that zones closer to the root are processed earlier.
			m_next_small_timer = time(NULL) + m_zone_index.get_32bit_chunk(5);
			// Star this zone.
			start_timer();

		}

	template<typename _uintX, class debugger>
		bool routing_zone<_uintX, debugger>::is_leaf() const
		{
			return (m_bucket != NULL);
		}

	template<typename _uintX, class debugger>
		bool routing_zone<_uintX, debugger>::can_split() const
		{
			if(m_level >= uint32_t(uint_type::digits))
				return false;
			if((m_zone_index < KK || m_level < KBASE) && m_bucket->get_size() == K)
				return true;
			return false;
		}

		// Returns all contacts from this zone tree that are no deeper than *depth* from the current zone.
	template<typename _uintX, class debugger>
		void routing_zone<_uintX, debugger>::top_depth(
				int depth,
				contact_list *list_result,
				bool empty_first)
		{
			if (is_leaf())
				m_bucket->get_all_nodes(list_result, empty_first);
			else if (depth <= 0)
				random_bucket(list_result, empty_first);
			else
			{
				m_subzones[0]->top_depth(depth - 1, list_result, empty_first);
				m_subzones[1]->top_depth(depth - 1, list_result, empty_first);
			}

		}

		// Returns the maximum depth of the tree as the number of edges of the longest path to a leaf.
	template<typename _uintX, class debugger>
		size_t routing_zone<_uintX, debugger>::get_max_depth() const
		{
			boost::mutex::scoped_lock lock(m_mutex);
			if (is_leaf())
				return 0;
			return 1 + max(m_subzones[0]->get_max_depth(), m_subzones[1]->get_max_depth());

		}

	template<typename _uintX, class debugger>
		void routing_zone<_uintX, debugger>::random_bucket(contact_list *list_result, bool empty_first)
		{
			boost::mutex::scoped_lock lock(m_mutex);
			if (is_leaf())
				m_bucket->get_all_nodes(list_result, empty_first);
			else
				m_subzones[rand()&1]->random_bucket(list_result, empty_first);

		}

	template<typename _uintX, class debugger>
		void routing_zone<_uintX, debugger>::split()
		{
			boost::mutex::scoped_lock lock(m_mutex);
			debugger::send(ROUTETABLE,
					"split zone level %d, index %s",
					m_level,
					m_zone_index.get_binary_string(true).c_str());
			stop_timer();

			m_subzones[0] = gen_subzone(0);
			m_subzones[1] = gen_subzone(1);

			contact_list list_nodes;
			m_bucket->get_all_nodes(&list_nodes);
			for (contact_list_const_iterator iter = list_nodes.begin();
					iter != list_nodes.end();
					++iter)
			{
				int superzone = (*iter)->get_distance().get_bit_number(m_level);
				m_subzones[superzone]->m_bucket->add_contact(*iter);
			}

			m_bucket->m_dont_delete_contacts = true;
			delete m_bucket;
			m_bucket = NULL;

		}	

	template<typename _uintX, class debugger>
		void routing_zone<_uintX, debugger>::start_timer()
		{
			time_t now = time(NULL);
			// Start filling the tree, closest buckets first.
			m_next_big_timer = now + 10;
			kad<uint_type, debugger>::add_event(this);
		}

	template<typename _uintX, class debugger>
		void routing_zone<_uintX, debugger>::stop_timer()
		{
			kad<uint_type, debugger>::remove_event(this);
		}
	
	template<typename _uintX, class debugger>
		void routing_zone<_uintX, debugger>::random_lookup()
		{
			// Look-up a random client in this zone
			uint_type prefix(m_zone_index);
			prefix.shift_left((uint_type::digits  - 1) - m_level);
			uint_type random(prefix, m_level);
			random._xor(m_kad_id);
			debugger::send(ROUTETABLE,
				"send random client lookup for %s",
				random.get_hex_string().c_str());
			session_manager<uint_type, debugger>::find_node(random, false);

		}

	template<typename _uintX, class debugger>
		class routing_zone<_uintX, debugger>* routing_zone<_uintX, debugger>::gen_subzone(int side)
		{
			uint_type new_index(m_zone_index);
			new_index.shift_left(1);
			if (side != 0)
				new_index.add(1);
			return new routing_zone(this, m_level + 1, new_index);

		}

	template<typename _uintX, class debugger>
		typename routing_zone<_uintX, debugger>::uint_type
		routing_zone<_uintX, debugger>::m_kad_id = uint_type(false) ;
}
#endif //ROUTETABLE_HPP
