/*
Copyright (C)2006 George Ang (gnap.an AT gmail.com)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.
 
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
 
You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/ 


#include <iostream>
#include <openssl/sha.h>
#include <kademlia/utils/uint160.hpp>

uint160::uint160()
{
	set_value((uint32_t)0);
}

uint160::uint160(bool bitfill)
{
	if( bitfill )
	{
		for (int i = 0; i < 20; i++)
			m_data[i] = (unsigned char)-1;
	}
	else
		set_value((uint32_t)0);
}

uint160::uint160(uint32_t value)
{
	set_value(value);
}

uint160::uint160(const uint160 &value, unsigned int count)
{
	unsigned int num_ulongs = count / 8;
	for( unsigned int index = 0; index < num_ulongs; index++)
		m_data[index] = value.m_data[index];
	// Copy the remaining bits
	for( unsigned int index = (8*num_ulongs); index < count; index++)
		set_bit_number(index, value.get_bit_number(index));
	// Pad with random bytes (Not seeding based on time to allow multiple different ones to be created in quick succession)
	for( unsigned int index = count; index < 160; index++)
		set_bit_number(index, (rand()%2));
}

uint160& uint160::set_value(const uint160 &value)
{
	for(int i = 0; i < 20; i++)
		m_data[i] = value.m_data[i];
	return *this;
}

uint160& uint160::set_value(uint32_t value)
{
	for(int i = 0; i < 16; i++)
		m_data[i] = 0;
	m_data[19] = (unsigned char)(value         & 0xFF);
	m_data[18] = (unsigned char)((value >>  8) & 0xFF);
	m_data[17] = (unsigned char)((value >> 16) & 0xFF);
	m_data[16] = (unsigned char)((value >> 24) & 0xFF);
	return *this;
}

uint160& uint160::set_value_be(const unsigned char *value)
{
	set_value((uint32_t)0);
	for (int index = 0; index < 20; index++)
		m_data[index] = (unsigned char)value[index];
	return *this;
}

uint160& uint160::set_value_random()
{
	unsigned char randombytes[20];
	for (int index = 0; index < 20; index++)
		randombytes[index] ^= rand();
	set_value_be( randombytes );
	return *this;
}

uint160& uint160::set_value_hex(const char *str)
{
	for(int i = 0; i < 5; ++i)
	{
		uint32_t val = 0x00000000;
		sscanf(str + i*8, "%8X", &val);
		set_32bit_chunk(i, val);
	}	
	return *this;
}

uint32_t uint160::get_32bit_chunk(int pos) const
{
	uint32_t value = 0;
	uint32_t start = (pos % 5) * 4;
	value |= ( (uint32_t)m_data[start    ] << 24 );
	value |= ( (uint32_t)m_data[start + 1] << 16 );
	value |= ( (uint32_t)m_data[start + 2] <<  8 );
	value |= ( (uint32_t)m_data[start + 3]       );
	return value;

}

void uint160::set_32bit_chunk(int pos, uint32_t val)
{
	int start = (pos % 5) * 4;
	m_data[start + 3] = (unsigned char)( val	 & 0xFF);
	m_data[start + 2] = (unsigned char)((val >>  8)	 & 0xFF);
	m_data[start + 1] = (unsigned char)((val >> 16)  & 0xFF);
	m_data[start    ] = (unsigned char)((val >> 24)  & 0xFF);
}

unsigned char uint160::get_8bit_chunk(int val) const
{
	return m_data[val];
}

unsigned int uint160::get_bit_number(unsigned int bit) const
{
	if (bit > 159)
		return 0;
	int long_num = bit / 8;
	int shift = 7 - (bit % 8);
	return ((m_data[long_num] >> shift) & 1);
}

uint160& uint160::set_bit_number(unsigned int bit, unsigned int value)
{
	int long_num = bit / 8;
	int shift = 7 - (bit % 8);
	m_data[long_num] |= (1 << shift);
	if (value == 0)
		m_data[long_num] ^= (1 << shift);
	return *this;
}

uint160& uint160::_xor(const uint160 &value)
{
	for ( int index = 0; index < 20; index++)
		m_data[index] ^= value.m_data[index];
	return *this;
}

uint160& uint160::_xor(const unsigned char *value)
{
	return _xor(uint160(false).set_value_be(value));
}

int uint160::log() const 
{
	const static char logtable[256] = {
	     	-1, 0, 1, 1, 2, 2, 2, 2, 3, 3, 3, 3, 3, 3, 3, 3,
		4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4,
		5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
		5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
		6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6,
		6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6,
		6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6,
		6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6,
		7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7,
		7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7,
		7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7,
		7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7,
		7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7,
		7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7,
		7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7,
		7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7,
	};
	int i;
	int l = 152;
	const unsigned char *op = m_data;
	for(i=0; i < 20; i++, op++) {
		if(*op != 0)
		{
			return l + logtable[*op];
		}
		l -= 8;
	}
	return -1; /* all bytes were zero */
}


std::string uint160::get_hex_string() const
{
	char str[41];
	for ( int index = 0; index < 20; index++)
	{
		::sprintf(str + (index*2), "%02X", m_data[index] & 0xFF);
	}
	return std::string(str);
}

std::string uint160::get_binary_string(bool trim) const
{
	

	char str[161];
	int bit, pos = 0;
	for ( int index = 0; index < 160; index++)
	{
		bit = get_bit_number(index);
		if ((!trim) || (bit != 0))
		{
			::sprintf(str + pos, "%d", bit);
			trim = false;
			pos++;
		}
	}
	if (pos == 0)
		str[0] = '0';
	str[pos + 1] = '\0';
	return std::string(str);
}

int uint160::compare_to(const uint160 &other) const
{
	for ( int index = 0; index < 20; index++)
	{
		if (m_data[index] < other.m_data[index])
			return -1;
		if (m_data[index] > other.m_data[index])
			return 1;
	}
	return 0;
}

int uint160::compare_to(uint32_t value) const
{
	return compare_to(uint160(value));
}
	
uint160& uint160::add(const uint160 &value)
{
	if (value == 0)
		return *this;
        uint64_t sum = 0;
	for (int index = 19; index >= 0; index--)
	{
		sum += m_data[index];
		sum += value.m_data[index];
		m_data[index] = (uint32_t)sum;
		sum = sum >> 8;
	}
	return *this;
}

uint160& uint160::add(uint32_t value)
{
	if ( value == 0)
		return *this;
	add(uint160(value));
	return *this;
}

uint160& uint160::subtract(const uint160 &value)
{
	if (value == 0)
		return *this;
	uint64_t sum = 0;
	for ( int index = 19; index >= 0; index--)
	{
		sum += m_data[index];
		sum -= value.m_data[index];
		m_data[index] = (uint32_t)sum;
		sum = sum >> 8;
	}
	return *this;
}

uint160& uint160::subtract(uint32_t value)
{
	if (value == 0)
		return *this;
	subtract(uint160(value));
	return *this;
}

uint160& uint160::shift_left(unsigned int bits)
{
	if ((bits == 0) || (compare_to(0) == 0))
		return *this;
	if (bits > 159)
	{
		set_value((uint32_t)0);
		return *this;
	}

	unsigned char result[] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
	int index_shift = (int)bits / 8;
	uint16_t shifted = 0;
	for (int index = 19; index >= index_shift; index--)
	{
		shifted += ((uint16_t)m_data[index]) << (bits % 8);
		result[index - index_shift] = (unsigned char) shifted;
		shifted = shifted >> 8;
	}
	for (int index = 0; index < 20; index++)
		m_data[index] = result[index];
	return *this;
}


const unsigned char* uint160::get_data() const
{
	return m_data;
}

unsigned char* uint160::c_str()
{
	return m_data;
}

void uint160::operator+= (const uint160 &value)
{
	add(value);
}

void uint160::operator-= (const uint160 &value)
{
	subtract(value);
}	

void uint160::operator= (const uint160 &value)
{
	set_value(value);
}

bool uint160::operator< (const uint160 &value) const
{
	return (compare_to(value) < 0);
}

bool uint160::operator> (const uint160 &value) const
{
	return (compare_to(value) > 0);
}

bool uint160::operator<= (const uint160 &value) const
{
	return (compare_to(value) >= 0);
}

bool uint160::operator>= (const uint160 &value) const
{
	return (compare_to(value) <= 0);
}

bool uint160::operator== (const uint160 &value) const
{
	return (compare_to(value) == 0);
}

bool uint160::operator!= (const uint160 &value) const
{
	return (compare_to(value) != 0);
}

void uint160::operator+= (uint32_t value) 
{
	add(value);
}

void uint160::operator-= (uint32_t value)
{
	subtract(value);
}

void uint160::operator= (uint32_t value)
{
	set_value(value);
}

bool uint160::operator< (uint32_t value) const
{
	return (compare_to(value) < 0);
}

bool uint160::operator> (uint32_t value) const
{
	return (compare_to(value) > 0);
}

bool uint160::operator<= (uint32_t value) const
{
	return (compare_to(value) <= 0);
}

bool uint160::operator>= (uint32_t value) const
{
	return (compare_to(value) >= 0);
}

bool uint160::operator== (uint32_t value) const
{
	return (compare_to(value) == 0);
}

bool uint160::operator!= (uint32_t value) const
{
	return (compare_to(value) != 0);
}

uint160 get_sha1_hash(std::string const& str)
{
	uint160 result;
	unsigned char hash[SHA_DIGEST_LENGTH];
	SHA1((unsigned char *)(str.c_str()), str.size(), hash);
	result.set_value_be(hash);
	return result;

};

